"""Functions for working with Datastores"""

import contextlib
import gzip
import hashlib
import json
import pathlib
import tarfile
import tempfile

import libcloud.storage
from libcloud.storage.types import ObjectDoesNotExistError, Provider

from oocli import config
from oocli.data import DataStore, Run


def _rclone_environment(datastore):
    """Environment configuration needed to connect to the given datastore"""
    match datastore.provider:
        case "scaleway":
            credentials = config.scaleway_credentials()
            return (
                "RCLONE_CONFIG_REMOTE_TYPE=s3 "
                f"RCLONE_CONFIG_REMOTE_ACCESS_KEY_ID={credentials['SCW_ACCESS_KEY']} "
                f"RCLONE_CONFIG_REMOTE_SECRET_ACCESS_KEY={credentials['SCW_SECRET_KEY']} "
                f"RCLONE_CONFIG_REMOTE_REGION={datastore.region} "
                f"RCLONE_CONFIG_REMOTE_ENDPOINT=s3.{datastore.region}.scw.cloud "
            )
        case "gcp":
            credentials = config.gcp_credentials()
            return (
                "RCLONE_CONFIG_REMOTE_TYPE=gcs "
                f"RCLONE_CONFIG_REMOTE_SERVICE_ACCOUNT_CREDENTIALS='{json.dumps(credentials)}' "
                "RCLONE_CONFIG_REMOTE_BUCKET_POLICY_ONLY=true "
            )
        case _:
            raise RuntimeError(
                f"{datastore.provider}' is not a supported datastore provider"
            )


def mount_command(datastore: DataStore, path: str, target: str):
    """Command to mount the datastore"""
    return (
        f"{_rclone_environment(datastore)} "
        "rclone"
        " --read-only"
        " --log-systemd mount"
        " --poll-interval 0"
        " --cache-dir /var/rclone"
        " --dir-cache-time 1h"
        " --vfs-cache-mode full"
        " --vfs-cache-min-free-space 5G"
        " --vfs-cache-max-age 1d"
        " --vfs-cache-poll-interval 30s"
        " --vfs-read-ahead 128M"
        f" remote:{pathlib.Path(datastore.bucket) / path}"
        f" {target}"
    )


def copy_source_command(run: Run, target: str):
    """Command to copy a run's source code to the target location"""
    if run.commit_sha is None:
        return ""

    source_tar = pathlib.Path(run.datastore.bucket) / f"{run.commit_sha}.tar.gz"
    return (
        f"{_rclone_environment(run.datastore)} "
        f"rclone cat remote:{source_tar} | tar xz -C {target}"
    )


def podman_args(run: Run):
    """Extra podman arguments needed for the given run"""
    args = [f"-v /mnt/{i.sha}:/{i.sha}:ro" for i in run.inputs]
    args.append("-v /mnt/output:/output")
    args.append("-v /mnt/workspace:/workspace")
    return " ".join(args)


def store_output_command(run: Run, path: str):
    """Command to store the run's output files"""
    return (
        f"{_rclone_environment(run.datastore)} "
        f"rclone sync {path} remote:{pathlib.Path(run.datastore.bucket) / run.sha} "
    )


def store_log_command(run: Run):
    """Command to store the run's logs"""
    log_file = f"{run.sha}.log.gz"
    return (
        "journalctl --output=json | gzip | "
        f"{_rclone_environment(run.datastore)} "
        f"rclone rcat remote:{pathlib.Path(run.datastore.bucket) / log_file} "
    )


def get_logs(run: Run):
    """Command to retreive the run's logs"""
    with tempfile.NamedTemporaryFile(suffix=".gz") as archive:
        try:
            get(run.datastore, f"{run.sha}.log.gz", archive.name)
        except ObjectDoesNotExistError:
            raise RuntimeError(f"No log found for {run.short_sha}") from None
        with gzip.open(archive) as f:
            return f.readlines()


def _get_bucket(datastore: DataStore):
    match datastore.provider:
        case "scaleway":
            Driver = libcloud.storage.providers.get_driver(datastore.provider)
            credentials = config.scaleway_credentials()
            driver = Driver(
                credentials["SCW_ACCESS_KEY"],
                credentials["SCW_SECRET_KEY"],
                region=datastore.region,
            )
        case "gcp":
            Driver = libcloud.storage.providers.get_driver(Provider.GOOGLE_STORAGE)
            credentials = config.gcp_credentials()
            driver = Driver(
                credentials["client_email"],
                credentials,
                project=credentials["project_id"],
            )
        case _:
            raise RuntimeError(
                f"{datastore.provider}' is not a supported datastore provider"
            )
    return driver.get_container(datastore.bucket)


def list_contents(datastore: DataStore, prefix: str):
    """List the datastore's files with the given prefix"""
    bucket = _get_bucket(datastore)
    return [i.name.removeprefix(prefix) for i in bucket.list_objects(prefix=prefix)]


def put(datastore: DataStore, path: str, target: str):
    """Store a file in the datastore"""
    bucket = _get_bucket(datastore)
    bucket.upload_object(path, target)


def get(datastore: DataStore, path: str, target: str):
    """Retreive a file from the datastore"""
    bucket = _get_bucket(datastore)
    bucket.get_object(path).download(target, overwrite_existing=True)


def exists(datastore: DataStore, path: str | None = None):
    """Does the datastore exits, and if a path is given, does it also exist"""
    try:
        bucket = _get_bucket(datastore)
    except libcloud.storage.types.ContainerDoesNotExistError:
        return False

    if path is not None:
        try:
            bucket.get_object(path)
        except libcloud.storage.types.ObjectDoesNotExistError:
            return False

    return True


def store_source(datastore: DataStore, repository, commit, *, allow_dirty=False):
    """Store source code in datastore

    Args:
        datastore: The datastore to store to
        repository: The Git repository holding the source
        commit: The commit to store, if None, HEAD is used
        allow_dirty: Allow and include changes in the working director
    """
    file_name = lambda s: f"{s}.tar.gz"
    with tempfile.NamedTemporaryFile(suffix=".tar.gz") as f:
        if commit is None and repository.is_dirty():
            with open(f.name, "wb") as fileobj:
                # set mtime and filename for reproducable archives
                with gzip.GzipFile(
                    fileobj=fileobj, mode="wb", mtime=0, filename=""
                ) as gzipobj:
                    with tarfile.open(fileobj=gzipobj, mode="w") as tarobj:
                        for source_file in repository.git.ls_files().splitlines():
                            with contextlib.suppress(FileNotFoundError):
                                tarobj.add(source_file)
            with open(f.name, "rb") as fileobj:
                digest = hashlib.file_digest(fileobj, "sha256")
            source = "dirty~" + digest.hexdigest()
            if not exists(datastore, file_name(source)):
                put(datastore, f.name, file_name(source))
        else:
            source = repository.commit("HEAD" if commit is None else commit).hexsha
            if not exists(datastore, file_name(source)):
                repository.git.archive(source, "--output", f.name)
                put(datastore, f.name, file_name(source))
    return source



"""Systemd services and scripts deployed on environments"""

import textwrap

from libcloud.compute.deployment import (
    Deployment,
    MultiStepDeployment,
    ScriptDeployment,
)

from oocli import config, datastores
from oocli.data import Run


class ToFileDeployment(Deployment):
    """libcloud Deployment of contents to a target file"""

    def __init__(self, contents, target):
        self.contents = self._get_string_value(
            argument_name="contents", argument_value=contents
        )
        self.target = target

    def run(self, node, client):
        client.put(self.target, contents=self.contents)
        return node

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return (
            f"<ToFileDeployment contents={self.contents[:15]}... target={self.target}>"
        )


def mount_inputs(inputs: list[Run]):
    """Deploy systemd services that mount the given inputs"""
    deployments = MultiStepDeployment()
    for i in inputs:
        location = f"/mnt/{i.sha}/"
        deployments.add(
            [
                ToFileDeployment(
                    textwrap.dedent(
                        f"""\
                        #!/usr/bin/env bash
                        {datastores.mount_command(i.datastore, i.sha, location)}
                        """
                    ),
                    f"/mount-{i.sha}.sh",
                ),
                ToFileDeployment(
                    textwrap.dedent(
                        f"""\
                        [Unit]
                        Description=rclone mount of {location}
                        After=o-o.service
                        Requires=o-o.service
                        Before=o-o-job.service
                        BindsTo=o-o-job.service
                        [Service]
                        Type=notify
                        NotifyAccess=all
                        Restart=no
                        ExecStartPre=mkdir -p {location}
                        ExecStart=bash /mount-{i.sha}.sh
                        ExecStop=/bin/fusermount -u {location}
                        [Install]
                        RequiredBy=o-o-job.service
                        """
                    ),
                    f"/etc/systemd/system/o-o-mount-{i.sha}.service",
                ),
            ]
        )
    return deployments


def oo_context(run: Run, *, key, setup="", teardown=""):
    """Deploy a systemd service that starts up and tears down the environment"""
    return MultiStepDeployment(
        [
            ToFileDeployment(
                textwrap.dedent(
                    f"""\
                    #!/usr/bin/env bash
                    export DEBIAN_FRONTEND=noninteractive
                    apt-get -y update && apt-get -y install curl podman
                    wget https://downloads.rclone.org/v1.71.1/rclone-v1.71.1-linux-amd64.deb
                    apt-get -y install ./rclone-v1.71.1-linux-amd64.deb
                    mkdir -p /mnt/output/
                    mkdir -p /mnt/workspace/
                    echo 1 > /exit_status
                    {datastores.copy_source_command(run, '/mnt/workspace/')}
                    {setup}
                    """
                ),
                "/setup.sh",
            ),
            ToFileDeployment(
                textwrap.dedent(
                    f"""\
                    #!/usr/bin/env bash
                    EXIT_STATUS=$(cat /exit_status)
                    echo o-o ExitStatus=$EXIT_STATUS

                    {datastores.store_output_command(run, '/mnt/output')}

                    curl -X PUT \
                        -H "X-API-Key: {key}" \
                        {config.settings().apiurl}/runs/{run.project}/{run.sha}/completed?exit_status=$EXIT_STATUS

                    {datastores.store_log_command(run)}

                    {teardown}
                    """
                ),
                "/teardown.sh",
            ),
            ToFileDeployment(
                textwrap.dedent(
                    """\
                    [Unit]
                    Description=setup and teardown o-o
                    Requires=multi-user.target
                    After=multi-user.target
                    BindsTo=o-o-job.service
                    [Service]
                    Type=oneshot
                    ExecStart=bash /setup.sh
                    Restart=no
                    RemainAfterExit=true
                    ExecStopPost=bash /teardown.sh
                    """
                ),
                "/etc/systemd/system/o-o.service",
            ),
        ]
    )


def job(*, command, image, podman_args=""):
    """Deploy a systemd service that runs a job"""
    return MultiStepDeployment(
        [
            ToFileDeployment(
                textwrap.dedent(
                    f"""\
                    [Unit]
                    Description=o-o job
                    After=o-o.service
                    Requires=o-o.service
                    OnFailure=o-o-failure.service
                    [Service]
                    Type=exec
                    Restart=no
                    ExecStart=podman run \
                        --entrypoint=/bin/sh \
                        --workdir=/workspace \
                        --quiet \
                        --init \
                        {podman_args} \
                        {image} \
                        -c {command}
                    ExecStopPost=bash -c 'echo $EXIT_STATUS > /exit_status'
                    """
                ),
                "/etc/systemd/system/o-o-job.service",
            ),
            ToFileDeployment(
                textwrap.dedent(
                    """\
                    [Unit]
                    Description=Stop o-o job on failure
                    [Service]
                    Type=oneshot
                    Restart=no
                    ExecStart=systemctl stop o-o-job
                    """
                ),
                "/etc/systemd/system/o-o-failure.service",
            ),
        ]
    )


def start():
    """Deploy a script that starts the job systemd service"""
    return ScriptDeployment(
        textwrap.dedent(
            """\
            #!/usr/bin/env bash
            systemctl daemon-reload
            systemctl enable /etc/systemd/system/o-o-mount-*
            systemctl start o-o-job
            """
        ),
        name="/startup.sh",
    )



"""Functions for working with projects"""

import requests

from oocli import config


def read_all():
    """Retreive all projects"""
    response = requests.get(
        f"{config.settings().apiurl}/projects",
        headers={"X-API-Key": config.settings().token},
    )
    response.raise_for_status()
    return response.json()



"""Adjustments and fixes to the libcloud package

* Scaleway: Handle `baremetal` no longer a VM property returned by API
* Scaleway: Add # of GPUs from API to NodeSize
* Scaleway: Switch API to zoned endpoints at `api.scaleway.com`
* Scaleway: Be sure to pass regions (aka, zone) to API in `create_node`, `destroy_node`,
    and `wait_until_running`.
* Google: Use credentials file when given to GoogleStorageDriver
"""

import json

import libcloud
import libcloud.compute.drivers.scaleway
import libcloud.compute.types
import libcloud.storage.drivers.google_storage
import libcloud.storage.types
from libcloud.common.base import ConnectionUserAndKey
from libcloud.common.types import ProviderError
from libcloud.compute.base import NodeLocation

libcloud.compute.ssh.LIBCLOUD_PARAMIKO_SHA2_BACKWARD_COMPATIBILITY = False

# ruff: noqa: UP031


class ScalewayConnection(libcloud.compute.drivers.scaleway.ScalewayConnection):
    def request(
        self,
        action,
        params=None,
        data=None,
        headers=None,
        method="GET",
        raw=False,
        stream=False,
        region=None,
    ):
        """Fixes: Switch to using zoned API endpoints"""
        if region is None:
            raise RuntimeError("region is required for request")
        self.host = "api.scaleway.com"
        if action.lstrip("/").split("/")[0] == "volumes":
            action = f"/block/v1/zones/{region}/{action}"
        else:
            action = f"/instance/v1/zones/{region}/{action}"
        return ConnectionUserAndKey.request(
            self, action, params, data, headers, method, raw, stream
        )


class ScalewayNodeDriver(libcloud.compute.drivers.scaleway.ScalewayNodeDriver):
    connectionCls = ScalewayConnection

    def _to_size(self, name, size, availability):
        """Fixes: Handle unavailable baremetal property and add # of GPUs"""
        baremetal_missing = "baremetal" not in size
        if baremetal_missing:
            size["baremetal"] = None

        node_size = super()._to_size(name, size, availability)
        node_size.extra["gpus"] = size["gpu"]

        if baremetal_missing:
            del node_size.extra["baremetal"]

        return node_size

    def create_node(
        self,
        name,
        size,
        image,
        ex_volumes=None,
        ex_tags=None,
        region=None,
    ):
        """Fixes: Add region (aka zone) to API requests, fix volume loop"""
        data = {
            "name": name,
            "organization": self.key,
            "image": image.id,
            "volumes": ex_volumes or {},
            "commercial_type": size.id,
            "tags": ex_tags or [],
        }

        allocate_space = image.extra.get("size", 50)
        for volume in data["volumes"].values():  # change from libcloud version
            allocate_space += libcloud.compute.drivers.scaleway._to_lib_size(
                volume["size"]
            )

        while allocate_space < size.disk:
            if size.disk - allocate_space > 150:
                bump = 150
            else:
                bump = size.disk - allocate_space

            vol_num = len(data["volumes"]) + 1
            data["volumes"][str(vol_num)] = {
                "name": "%s-%d" % (name, vol_num),
                "organization": self.key,
                "size": libcloud.compute.drivers.scaleway._to_api_size(bump),
                "volume_type": "l_ssd",
            }
            allocate_space += bump

        if allocate_space > size.extra.get("max_disk", size.disk):
            range = (
                "of %dGB" % size.disk
                if size.extra.get("max_disk", size.disk) == size.disk
                else "between %dGB and %dGB"
                % (size.extra.get("max_disk", size.disk), size.disk)
            )
            raise ProviderError(
                value=(
                    "%s only supports a total volume size %s; tried %dGB"
                    % (size.id, range, allocate_space)
                ),
                http_code=400,
                driver=self,
            )

        response = self.connection.request(
            "/servers", data=json.dumps(data), region=region, method="POST"
        )
        server = response.object["server"]
        node = self._to_node(server)
        node.extra["region"] = (
            region.id if isinstance(region, NodeLocation) else region
        ) or "par1"

        # Scaleway doesn't start servers by default, let's do it
        self._action(node.id, "poweron", region=region)  # change from libcloud version

        return node

    def destroy_node(self, node):
        """Fixes: Add region (aka zone) to API request"""
        region = node.extra["region"]
        return self._action(node.id, "terminate", region=region)

    def wait_until_running(
        self,
        nodes,  # type: List[Node]
        ex_list_nodes_kwargs=None,  # type: Optional[Dict]
        **kwargs,
    ):
        """Fixes: Add region (aka zone) to API request"""
        region = nodes[0].extra["region"]
        assert all(n.extra["region"] == region for n in nodes)
        ex_list_nodes_kwargs = ex_list_nodes_kwargs or {}
        ex_list_nodes_kwargs["region"] = region
        return super().wait_until_running(
            nodes, ex_list_nodes_kwargs=ex_list_nodes_kwargs, **kwargs
        )


del libcloud.compute.providers.DRIVERS[libcloud.compute.types.Provider.SCALEWAY]
libcloud.compute.providers.set_driver(
    libcloud.compute.types.Provider.SCALEWAY,
    "oocli.libcloudfixes",
    "ScalewayNodeDriver",
)


class GoogleStorageDriver(libcloud.storage.drivers.google_storage.GoogleStorageDriver):
    """Fixes: Use credientials file when given to GoogleStorageDriver"""

    def __init__(self, key, secret=None, credential_file=None, **kwargs):
        self.credential_file = credential_file
        super().__init__(key, secret=secret, credential_file=credential_file, **kwargs)

    def _ex_connection_class_kwargs(self):
        return {
            "credential_file": self.credential_file,
        }


del libcloud.storage.providers.DRIVERS[libcloud.storage.types.Provider.GOOGLE_STORAGE]
libcloud.storage.providers.set_driver(
    libcloud.storage.types.Provider.GOOGLE_STORAGE,
    "oocli.libcloudfixes",
    "GoogleStorageDriver",
)



"""Common data classes"""

from pydantic import BaseModel, ByteSize, constr
from pydantic_extra_types.pendulum_dt import DateTime


class DataStore(BaseModel):
    provider: constr(to_lower=True)
    bucket: str
    region: str | None = None
    default: bool = False


class Environment(BaseModel):
    provider: constr(to_lower=True)
    image: str
    machinetype: str
    region: str
    default: bool = False
    size: ByteSize = ByteSize(20_000_000_000)


class LogMessage(BaseModel):
    message: str
    timestamp: DateTime


class Run(BaseModel):
    command: list[str]
    commit_sha: str | None = None
    creator: str
    datastore: DataStore
    environment: Environment
    message: str
    project: str
    started: DateTime
    ended: DateTime | None = None
    exit_status: int | None = None
    sha: str | None = None
    tags: list[str]
    inputs: list["Run"] | None = None

    @property
    def short_sha(self):
        return self.sha[:10]



"""Configuration and settings management"""

import functools
import json
import os
import pathlib

import yaml
from pydantic import FilePath
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)

from oocli.data import DataStore, Environment

CONFIG_PATH = pathlib.Path.home() / ".config" / "o-o"
CONFIG_FILE = CONFIG_PATH / "settings.yaml"

DEFAULT_HOST = "https://o-o.tools"


class Settings(BaseSettings):
    """o-o settings

    Settings are loaded from both an `.ooconfig` file in the current working
    directory and from `$HOME/.config/o-o/settings.yaml`. Preference is given
    to settings in `.ooconfig`.
    """

    host: str
    sshkey: FilePath
    project: str
    token: str
    datastores: dict[str, DataStore]
    environments: dict[str, Environment]
    sourcecode: bool = False

    model_config = SettingsConfigDict(yaml_file=[CONFIG_FILE, ".ooconfig"])

    @property
    def apiurl(self):
        return f"{self.host}/api/v0"

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (YamlConfigSettingsSource(settings_cls),)


@functools.cache
def settings() -> Settings:
    """Cached interface to settings"""
    return Settings()


def get_setting(setting: str, value: str | None):
    """Get a configured setting, or if value is None, return the default"""
    configurations = getattr(settings(), setting)
    if value is None:
        if len(configurations) == 1:
            return next(iter(configurations.values()))
        try:
            return next(e for e in configurations.values() if e.default)
        except StopIteration:
            raise RuntimeError(
                f"no default configured, must select provide one of {list(configurations.keys())}"
            ) from None
    if value not in configurations:
        raise RuntimeError(f"'{value}' is not one of {list(configurations.keys())}")
    return configurations[value]


def write(**kwargs):
    """Write settings to config file location"""
    if CONFIG_FILE.exists():
        kwargs = yaml.safe_load(CONFIG_FILE.read_text()) | kwargs
    CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
    CONFIG_FILE.write_text(yaml.dump(kwargs))


@functools.cache
def gcp_credentials():
    """Retreive Google Cloud credentials"""
    credentials_file = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS", None)
    if credentials_file is None:
        raise RuntimeError(
            "GOOGLE_APPLICATION_CREDENTIALS environment variable not set"
        )
    credentials_file = pathlib.Path(credentials_file)
    if not credentials_file.exists():
        raise RuntimeError(
            f"Google application credentials file not found ({credentials_file.as_posix()}). "
        )
    return json.loads(credentials_file.read_text())


@functools.cache
def scaleway_credentials():
    """Retreive Scaleway credentials"""
    required_variables = [
        "SCW_ACCESS_KEY",
        "SCW_SECRET_KEY",
        "SCW_DEFAULT_ORGANIZATION_ID",
    ]
    missing_variables = [v for v in required_variables if v not in os.environ]
    if missing_variables:
        raise RuntimeError(
            f"Scaleway environment variables not set: {', '.join(missing_variables)}"
        )

    return {v: os.environ[v] for v in required_variables}



"""o-o command line interface"""

import contextlib
import functools
import io
import pathlib
import re
import tarfile
import tempfile

import asciidag.graph
import asciidag.node
import click
import dotenv
import git
import pendulum
import typer
from rich import padding
from rich.console import Console
from rich.markup import escape
from rich.prompt import Prompt
from rich.table import Table
from rich.text import Text
from rich.theme import Theme

from oocli import clitools, config, data, datastores, projects, runs, tags

app = typer.Typer(
    add_completion=False,
    rich_markup_mode="markdown",
    pretty_exceptions_show_locals=False,
)

console = Console(
    highlight=False,
    theme=Theme(
        {
            "info": "dim cyan",
            "warning": "magenta",
            "danger": "bold red",
            "sha": "yellow",
        }
    ),
)
print = console.print
dotenv.load_dotenv(dotenv_path=dotenv.find_dotenv(usecwd=True))


def command(*args, **kwargs):
    """Wrap the typer command for uncaught exception handling"""

    def decorator(func):
        @app.command(*args, **kwargs)
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except (typer.BadParameter, typer.Exit):
                raise
            except Exception as e:
                print(f"[danger]{e.__class__.__name__}:[/danger] {e}")
                raise typer.Exit(1) from None

        return wrapper

    return decorator


@command("cp")
def cp_command(
    sources: list[str] = typer.Argument(..., hidden=True, metavar="SOURCE..."),
    destination: str = typer.Argument(..., hidden=True, metavar="DESTINATION"),
    datastore: str | None = typer.Option(
        None,
        "--datastore",
        "-d",
        show_default=False,
        help="Override default datastore.",
        rich_help_panel="Options for uploads",
    ),
    message: str | None = typer.Option(
        None,
        "--message",
        "-m",
        show_default=False,
        help="A description of the upload.",
        rich_help_panel="Options for uploads",
    ),
    project: str | None = typer.Option(
        None,
        "--project",
        "-p",
        help="Override configured project.",
        show_default=False,
        callback=clitools.value_or_config_default("project"),
    ),
    copy_tags: list[str] = typer.Option(
        [],
        "--tag",
        "-t",
        help="Tag the upload with the given tag(s).",
        rich_help_panel="Options for uploads",
    ),
):
    """Copy files to and from datastores

    Copy SOURCE to DESTINATION, or multiple SOURCE(s) to DESTINATION directory,
    where SOURCE and/or DESTINATION are o-o datastore locations. For example, to
    upload a file to o-o:

        $ o-o cp --message "Importing data" data.txt o://output/

    Uploading creates a dummy run where the uploaded files are ouputs to be
    used as inputs to other runs.

        $ o-o run --list
        ...
        fojaopi1t5 Importing data

    Outputs from any run (including our example upload above) can also be
    downloaded:

        $ o-o cp o://fojaopi1t5/data.txt .
    """
    if len(sources) > 1 and not destination.endswith("/"):
        print(
            "[warning]Destination must be a directory with multiple sources[/warning]"
        )
        raise typer.Exit(1)

    oo_path_pattern = re.compile(r"^o://(?P<sha>\w+)/(?P<path>\S*)")
    match = oo_path_pattern.match(destination)
    if match:
        try:
            datastore = config.get_setting("datastores", datastore)
        except RuntimeError as e:
            raise typer.BadParameter(str(e), param_hint=["--datastore", "-d"]) from None

        if match["sha"] != "output":
            print(
                f"[warning]Cannot write to {destination}, only o://output/ and local destinations are supported[/warning]"
            )
            raise typer.Exit(1)

        if message is None:
            message = click.edit(
                "Copying files into datastore\n"
                "# Please enter a message that describes this upload. Lines starting\n"
                "# with '#' will be ignored, and an empty message aborts this run.\n"
            )
            if message is not None:
                message = "\n".join(
                    line for line in message.splitlines() if not line.startswith("#")
                )
            if not message:
                print("[warning]Aborting run due to empty message.[/warning]")
                raise typer.Exit(1)

        run, _ = runs.create(
            command=["cp"] + sources + [destination],
            commit_sha=None,
            datastore=datastore,
            environment=data.Environment(
                provider="local", image="None", machinetype="None", region="None"
            ),
            message=message,
            project=project,
            tags=copy_tags,
        )

        try:
            destination_path = pathlib.Path(run.sha) / match["path"]
            destination_is_dir = match["path"].endswith("/") or match["path"] == ""

            with clitools.TaskDisplay(console) as tasks:
                for i, source in enumerate(sources):
                    tasks.set_current(
                        f"[{i+1}/{len(sources)}] copying {source} -> {destination}"
                    )
                    match = oo_path_pattern.match(source)
                    if match:
                        print(
                            f"[warning]Copying between datastores ({source} -> {destination}) not supported[/warning]"
                        )
                        raise typer.Exit(1)
                    else:
                        source_path = pathlib.Path(source)
                        if destination_is_dir:
                            datastores.put(
                                datastore,
                                source_path.as_posix(),
                                (destination_path / source_path.name).as_posix(),
                            )
                        else:
                            datastores.put(
                                datastore,
                                source_path.as_posix(),
                                destination_path.as_posix(),
                            )
        except KeyboardInterrupt:
            runs.completed(run, exit_status=130)
            raise
        except Exception:
            runs.completed(run, exit_status=1)
            raise
        else:
            runs.completed(run, exit_status=0)

    else:
        if datastore is not None:
            print("[info]Datastore provided, but is not used for downloads.[/info]")
        if message is not None:
            print("[info]Message provided, but is not used for downloads.[/info]")
        if copy_tags:
            print("[info]Tag provided, but is not used for downloads.[/info]")
        destination_path = pathlib.Path(destination)
        with clitools.TaskDisplay(console) as tasks:
            for i, source in enumerate(sources):
                tasks.set_current(
                    f"[{i+1}/{len(sources)}] copying {source} -> {destination}"
                )
                match = oo_path_pattern.match(source)
                if match:
                    source_run = runs.read(project, match["sha"])
                    source_path = pathlib.Path(source_run.sha) / match["path"]
                    if destination_path.is_dir():
                        datastores.get(
                            source_run.datastore,
                            source_path.as_posix(),
                            (destination_path / source_path.name).as_posix(),
                        )
                    else:
                        datastores.get(
                            source_run.datastore,
                            source_path.as_posix(),
                            destination_path.as_posix(),
                        )
                else:
                    print(
                        f"[warning]Copying from local file to local file ({source} -> {destination}) not supported[/warning]"
                    )
                    raise typer.Exit(1)


@command("diff")
def diff_command(
    left: str = typer.Argument(..., hidden=True, metavar="RUN"),
    right: str = typer.Argument(..., hidden=True, metavar="RUN"),
    project: str | None = typer.Option(
        None,
        "--project",
        "-p",
        help="Override configured project.",
        show_default=False,
        callback=clitools.value_or_config_default("project"),
    ),
):
    """Show code changes between runs"""
    left, right = [runs.read(project, r) for r in [left, right]]
    runs_without_source = [r for r in [left, right] if r.commit_sha is None]
    if runs_without_source:
        raise typer.BadParameter(
            f"{runs_without_source} where run without source files"
        )

    if left.commit_sha == right.commit_sha:
        return

    with tempfile.TemporaryDirectory() as directory:
        directory = pathlib.Path(directory)

        with clitools.TaskDisplay(console) as tasks:
            for run in [left, right]:
                tasks.set_current(f"Downlading {run.short_sha} source...")
                source_archive = f"{run.commit_sha}.tar.gz"
                datastores.get(
                    run.datastore,
                    source_archive,
                    directory / source_archive,
                )
                with tarfile.open(directory / source_archive) as tar:
                    tar.extractall(directory / run.short_sha)

        with contextlib.chdir(directory):
            clitools.print_diffs(
                pathlib.Path(left.short_sha),
                pathlib.Path(right.short_sha),
                print_fn=console.print,
            )


@command("log")
def log_command(
    run: str = typer.Argument(..., hidden=True),
    max_lines: int = typer.Option(200, help="Maximum number of lines."),
    show_all: bool = typer.Option(
        False,
        "--all",
        help="Show all system logged messages, not just from the job.",
    ),
    timestamps: bool = typer.Option(
        False,
        "--timestamps",
        help="Show timestamps in output.",
    ),
    project: str | None = typer.Option(
        None,
        "--project",
        "-p",
        help="Override configured project.",
        show_default=False,
        callback=clitools.value_or_config_default("project"),
    ),
):
    """Show run output"""
    try:
        for log in runs.connect(
            runs.read(project, run), max_lines=max_lines, show_all=show_all
        ):
            message = escape(log.message)
            if timestamps:
                timestamp = log.timestamp.in_tz(pendulum.local_timezone())
                timestamp = timestamp.format("L LTS")
                message = f"[dim]{timestamp}[/dim] {message}"
            print(message, highlight=False)
    except runs.RunFinished as exception:
        raise typer.Exit(exception.exit_code) from None


@command("login")
def login_command(
    host: str = typer.Option(
        config.DEFAULT_HOST,
        "--host",
        "-h",
        help="Override default host.",
    ),
):
    """Login to o-o"""
    host = host.rstrip("/")
    print(f"Visit {host}/activate")
    token = Prompt.ask("Paste token", console=console, password=True)
    while True:
        sshkey = Prompt.ask(
            "Enter ssh key location",
            console=console,
            default="~/.ssh/id_ed25519",
        )
        sshkey = pathlib.Path(sshkey).expanduser()
        if not sshkey.exists():
            print(f"[danger]ssh key {sshkey} not found")
        else:
            break
    config.write(token=token, host=host, sshkey=sshkey.as_posix())


@command("project")
def project_command():
    """List projects"""
    for project in projects.read_all():
        print(project)


@command("run")
def run_command(
    command: list[str] = typer.Argument(
        [],
        hidden=True,
        metavar="-- COMMAND [ARGS...]",
        show_default=False,
    ),
    list_runs: bool = typer.Option(
        False,
        "--list",
        "-l",
        help="List runs and exit (default if no COMMAND is given).",
    ),
    datastore: str | None = typer.Option(
        None,
        "--datastore",
        "-d",
        show_default=False,
        help="Override default datastore.",
    ),
    environment: str | None = typer.Option(
        None,
        "--environment",
        "-e",
        show_default=False,
        help="Override default environment.",
    ),
    message: str | None = typer.Option(
        None,
        "--message",
        "-m",
        show_default=False,
        help="A description of the run.",
    ),
    project: str | None = typer.Option(
        None,
        "--project",
        "-p",
        help="Override configured project.",
        show_default=False,
        callback=clitools.value_or_config_default("project"),
    ),
    run_tags: list[str] = typer.Option(
        [],
        "--tag",
        "-t",
        help="Tag the run with the given tag(s).",
    ),
    timestamps: bool = typer.Option(
        False,
        "--timestamps",
        help="Show timestamps in output.",
    ),
    sourcecode: bool = typer.Option(
        None,
        help="Upload Git repository source code to run environment (will override configuration).",
        callback=clitools.value_or_config_default("sourcecode"),
        rich_help_panel="Git source code options",
    ),
    commit: str | None = typer.Option(
        None,
        "--commit",
        "-c",
        show_default="HEAD",
        metavar="REF",
        help="Upload source with this reference (implies --sourcecode).",
        rich_help_panel="Git source code options",
    ),
    allow_dirty: bool = typer.Option(
        False,
        "--dirty",
        help="Upload source with changes in your working tree (implies --sourcecode).",
        rich_help_panel="Git source code options",
    ),
):
    """Create or list run(s)"""
    if list_runs or not command:
        for run in runs.read_all(project):
            print(clitools.one_line(run))
        raise typer.Exit(0)

    try:
        datastore = config.get_setting("datastores", datastore)
    except RuntimeError as e:
        raise typer.BadParameter(str(e), param_hint=["--datastore", "-d"]) from None
    try:
        environment = config.get_setting("environments", environment)
    except RuntimeError as e:
        raise typer.BadParameter(str(e), param_hint=["--environment", "-e"]) from None

    if message is None:
        message = click.edit(
            "\n"
            "# Please enter a message that describes this run. Lines starting\n"
            "# with '#' will be ignored, and an empty message aborts this run.\n"
        )
        if message is not None:
            message = "\n".join(
                line for line in message.splitlines() if not line.startswith("#")
            )
        if not message:
            print("[warning]Aborting run due to empty message.[/warning]")
            raise typer.Exit(1)

    if not datastores.exists(datastore):
        region_info = f" in '{datastore.region}'" if datastore.region else ""
        print(
            f"[warning]Invalid datastore[/warning]: {datastore.provider} "
            f"bucket '{datastore.bucket}' does not exists{region_info}, "
            f"please create with provider"
        )
        raise typer.Exit(1)

    with clitools.TaskDisplay(console=console) as tasks:
        source = None
        if sourcecode or commit or allow_dirty:
            tasks.set_current(
                f"Packaging git commit {commit or 'HEAD'}"
                f"{' with working directory changes' if allow_dirty else ''}...",
            )
            try:
                repo = git.Repo("./")
                if commit is not None and allow_dirty:
                    raise typer.BadParameter(
                        "--commit cannot be used with --dirty.",
                    )
                if commit is None and repo.is_dirty() and not allow_dirty:
                    raise typer.BadParameter(
                        "You have uncommitted changes, either commit changes or use --dirty / --commit"
                    )
                source = datastores.store_source(
                    datastore, repo, commit, allow_dirty=allow_dirty
                )
            except git.InvalidGitRepositoryError:
                param_hint = []
                if sourcecode:
                    param_hint.append("--sourcecode")
                if commit:
                    param_hint.append("--commit")
                if allow_dirty:
                    param_hint.append("--dirty")
                raise typer.BadParameter(
                    "connot be used, not in Git repositoriy",
                    param_hint=param_hint,
                ) from None
            except git.BadName:
                if commit is not None:
                    raise typer.BadParameter(
                        f"ref '{commit}' not found.",
                        param_hint=["--commit", "-c"],
                    ) from None
                raise

        run = None
        try:
            tasks.set_current("Creating run...")
            run, key = runs.create(
                command=command,
                commit_sha=source,
                datastore=datastore,
                environment=environment,
                message=message,
                project=project,
                tags=run_tags,
            )
            if run.inputs:
                tasks.set_current("Waiting for input(s) to complete...")
                runs.wait_for_inputs(run)
            tasks.set_current(f"Starting {environment.provider} environment...")
            runs.start(run, key)
        except KeyboardInterrupt:
            tasks.set_current(
                f"Run cancelled, stopping {environment.provider} environment..."
            )
            if run is not None:
                try:
                    runs.kill(run)
                except Exception as e:
                    print(f"[warning]Error occured while killing node: {e}[/warning]")
            return
        except:
            tasks.set_current(
                f"Error occurred, stopping {environment.provider} environment..."
            )
            if run is not None:
                try:
                    runs.kill(run)
                except Exception as e:
                    print(f"[warning]Error occured while killing node: {e}[/warning]")
            raise

        try:
            tasks.set_current(f"Running {run.short_sha}...")
            log_command(
                run.sha,
                max_lines=None,
                show_all=False,
                timestamps=timestamps,
                project=project,
            )
        except KeyboardInterrupt:
            print(
                f"[info]Exiting, but job is still running, reconnect with 'o-o log {run.short_sha}'[/info]"
            )
            return


@command("show")
def show_command(
    run: str = typer.Argument(..., hidden=True),
    extra: bool = typer.Option(
        False,
        "--extra",
        "-e",
        help="Show extra info (full command, datastore, environment).",
    ),
    inputs: bool = typer.Option(False, "--inputs", "-i", help="Show inputs."),
    outputs: bool = typer.Option(False, "--outputs", "-o", help="Show outputs."),
    project: str | None = typer.Option(
        None,
        "--project",
        "-p",
        help="Override configured project.",
        show_default=False,
        callback=clitools.value_or_config_default("project"),
    ),
):
    """Show run details"""
    run = runs.read(project, run)
    print(
        f"[sha]Run {run.sha}",
        f"[sha]([bold]{', '.join(run.tags)}[/bold])" if run.tags else "",
    )

    grid = Table.grid()
    grid.add_column()
    grid.add_column()
    grid.add_row("Creator: ", f"{run.creator}")
    if run.commit_sha is not None:
        grid.add_row("Commit: ", f"{run.commit_sha}")
    grid.add_row(
        "Started: ",
        f"{run.started.in_tz(pendulum.local_timezone()).to_rss_string()}",
    )
    if run.ended is not None:
        msg = f"{run.ended.in_tz(pendulum.local_timezone()).to_rss_string()}"
        if run.exit_status != 0:
            msg = msg + f" [red][{run.exit_status}][/red]"
        grid.add_row("Ended: ", msg)
    grid.add_row("Command: ", Text(" ".join(run.command), no_wrap=not extra))
    if extra:
        grid.add_row(
            "Environment: ",
            f"provider={run.environment.provider} machinetype={run.environment.machinetype} image={run.environment.image}",
        )
        grid.add_row(
            "Datastore: ",
            f"provider={run.datastore.provider} bucket={run.datastore.bucket}",
        )
    print(grid, highlight=False)
    print(padding.Padding(run.message, pad=(1, 0, 0, 4)))
    print()

    if inputs:
        node_table = {}

        def _graph_inputs(inputs):
            result = []
            for i in inputs:
                if i.sha in node_table:
                    result.append(node_table[i.sha])
                else:
                    new_node = asciidag.node.Node(
                        clitools.one_line(i),
                        parents=_graph_inputs(i.inputs),
                    )
                    result.append(new_node)
                    node_table[i.sha] = new_node
            return result

        print("Inputs:")
        if run.inputs:
            s = io.StringIO()
            graph = asciidag.graph.Graph(s)
            graph.show_nodes(
                [asciidag.node.Node("", parents=_graph_inputs(run.inputs))]
            )
            for line in s.getvalue().splitlines()[1:]:
                print("  " + re.sub(r"\* ", "o ", line))
        print()

    if outputs:
        print("Output files:")
        for o in runs.outputs(run):
            print(f"  {o}")
        print()


@command("stop")
def stop_command(
    run: str = typer.Argument(..., hidden=True),
    kill: bool = typer.Option(False, "--kill", "-k", help="Terminate the environment."),
    project: str | None = typer.Option(
        None,
        "--project",
        "-p",
        help="Override configured project.",
        show_default=False,
        callback=clitools.value_or_config_default("project"),
    ),
):
    """Stop a run"""
    run = runs.read(project, run)
    with clitools.TaskDisplay(console) as tasks:
        if kill:
            tasks.set_current(f"Killing {run.short_sha}...")
            runs.kill(run)
        else:
            tasks.set_current(f"Stopping {run.short_sha}...")
            runs.stop(run)


@command("tag")
def tag_command(
    name: str = typer.Argument(default=None, hidden=True, metavar="NAME"),
    run: str = typer.Argument(default=None, hidden=True, metavar="RUN"),
    list_tags: bool = typer.Option(
        False,
        "--list",
        "-l",
        help="List tags and exit (default if NAME and RUN is not given).",
    ),
    project: str | None = typer.Option(
        None,
        "--project",
        "-p",
        help="Override configured project.",
        show_default=False,
        callback=clitools.value_or_config_default("project"),
    ),
):
    """Create or list run tags"""
    if list_tags or (not name and not run):
        for tag in tags.read_all(project):
            print(tag)
        return

    if run and name:
        tags.create(name, run, project=project)
    else:
        raise typer.BadParameter(
            "NAME and RUN required when creating a tag",
        )


if __name__ == "__main__":
    app()



"""Tools for the command line interface"""

import difflib
import filecmp
import pathlib
import tempfile

import rich.progress
from rich.syntax import Syntax

from oocli import config, data


def value_or_config_default(setting: str):
    """Typer callback that returns configured setting if None"""

    def callback(value):
        return value if value is not None else getattr(config.settings(), setting)

    return callback


def one_line(run: data.Run):
    """Return run information formatted in one line of text"""
    line = f"[sha]{run.short_sha}[/sha]"
    if run.tags:
        line = line + f" [sha]([bold]{', '.join(run.tags)}[/bold])[/sha]"
    if run.ended is None:
        line = line + "[green]"
    elif run.exit_status != 0:
        line = line + "[red]"

    title, *_ = run.message.splitlines()
    return line + f" {title}"


def print_diffs(
    left: pathlib.Path,
    right: pathlib.Path,
    /,
    print_fn=print,
):
    """Print diffs between given directories"""

    def _unified_diff(left, right, /):
        udiff = difflib.unified_diff(
            "" if left is None else left.read_text().splitlines(keepends=True),
            "" if right is None else right.read_text().splitlines(keepends=True),
            "/dev/null" if left is None else str(left),
            "/dev/null" if right is None else str(right),
        )
        return "".join(udiff)

    with tempfile.TemporaryDirectory() as empty_dir:
        empty_dir = pathlib.Path(empty_dir)

        def _diffs(left, right, /):
            diffs = filecmp.dircmp(left, right)
            for diff in diffs.diff_files:
                yield _unified_diff(left / diff, right / diff)

            for diff in diffs.left_only:
                if (left / diff).is_dir():
                    yield from _diffs(left / diff, empty_dir)
                else:
                    yield _unified_diff(left / diff, None)

            for diff in diffs.right_only:
                if (right / diff).is_dir():
                    yield from _diffs(empty_dir, right / diff)
                else:
                    yield _unified_diff(None, right / diff)

            for subdir in diffs.common_dirs:
                yield from _diffs(left / subdir, right / subdir)

        for diff in _diffs(left, right):
            print_fn(Syntax("".join(diff), lexer="diff"))


class TaskDisplay:
    """Progress spinner"""

    def __init__(self, console=None, disable=False):
        self._progress = rich.progress.Progress(
            rich.progress.SpinnerColumn(),
            rich.progress.TextColumn("[dim yellow]{task.description}"),
            transient=True,
            console=console,
            disable=disable,
        )
        self._task = None

    def __enter__(self):
        self._progress.__enter__()
        return self

    def __exit__(self, *args):
        self._progress.__exit__(*args)

    def set_current(self, description):
        if self._task is not None:
            self._progress.update(self._task, visible=False)
        self._task = self._progress.add_task(description, total=None)



"""Functions for working with tags"""

import requests

from oocli import config


def create(name: str, run_sha: str, *, project: str):
    """Create or update a tag"""
    response = requests.put(
        f"{config.settings().apiurl}/tags/{project}/{name}",
        headers={"X-API-Key": config.settings().token},
        params={"sha": run_sha},
    )
    if response.status_code == 422:
        raise RuntimeError(response.text)
    response.raise_for_status()


def read_all(project: str):
    """Retreive all tags in the project"""
    response = requests.get(
        f"{config.settings().apiurl}/tags/{project}",
        headers={"X-API-Key": config.settings().token},
    )
    response.raise_for_status()
    return response.json()



"""Functions for working with Runs"""

import collections
import json
import re
import shlex
import textwrap
import time

import paramiko
import pendulum
import requests
from libcloud.compute.deployment import MultiStepDeployment

from oocli import config, datastores, services
from oocli.data import Environment, LogMessage, Run
from oocli.libcloudfixes import libcloud


def get_driver(environment: Environment):
    """Get libcloud driver for environment"""
    match environment.provider:
        case "gcp":
            Driver = libcloud.compute.providers.get_driver(
                libcloud.compute.types.Provider.GCE
            )
            credentials = config.gcp_credentials()
            driver = Driver(
                credentials["client_email"],
                credentials,
                project=credentials["project_id"],
                datacenter=environment.region,
            )
        case libcloud.compute.types.Provider.SCALEWAY:
            Driver = libcloud.compute.providers.get_driver(environment.provider)
            credentials = config.scaleway_credentials()
            driver = Driver(
                credentials["SCW_DEFAULT_ORGANIZATION_ID"],
                credentials["SCW_SECRET_KEY"],
            )
        case _:
            raise RuntimeError(
                f"'{environment.provider}' is not a supported environment provider"
            )
    return driver


def safe_command(run: Run):
    """Get a properly escaped command for the given run"""
    _OO_PATH_PATTERN = re.compile(r"(?P<space>\s)o://(?P<input>\w+)/")

    def repl(m):
        match m["input"]:
            case "output":
                return f"{m['space']}/output/"
            case i:
                i = read(run.project, i)
                return f"{m['space']}/{i.sha}/"

    return shlex.quote(_OO_PATH_PATTERN.sub(repl, " ".join(run.command)))


def create(*, project, environment, datastore, **kwargs):
    """Create a new run"""
    assert "sha" not in kwargs
    assert "started" not in kwargs
    assert "ended" not in kwargs
    assert "error" not in kwargs
    assert "inputs" not in kwargs

    data = {
        "project": project,
        "environment": environment.model_dump(),
        "datastore": datastore.model_dump(),
        **kwargs,
    }
    response = requests.post(
        f"{config.settings().apiurl}/runs/{project}",
        data=json.dumps(data),
        headers={"X-API-Key": config.settings().token},
    )
    if response.status_code == 422:
        raise RuntimeError(response.text)
    response.raise_for_status()
    sha = response.json()["sha"]
    key = response.json()["key"]
    return read(project, sha), key


def start(run, key):
    """Start a run"""
    match run.environment.provider:
        case "gcp":
            _start_gce(run, key)
        case libcloud.compute.types.Provider.SCALEWAY:
            _start_scaleway(run, key)
        case _:
            raise RuntimeError(
                f"'{run.environment.provider}' is not a supported environment provider"
            )


def _start_gce(run: Run, key: str):
    driver = get_driver(run.environment)
    private_key = config.settings().sshkey
    public_key = private_key.with_name(f"{private_key.name}.pub")
    metadata = {
        "items": [
            {
                "key": "ssh-keys",
                "value": f"root: {public_key.read_text().strip()}",
            },
        ],
    }

    sizes = driver.list_sizes()
    size = next((s for s in sizes if s.name == run.environment.machinetype), None)
    if size is None:
        raise RuntimeError(
            f"{run.environment.machinetype} not available in {run.environment.region}"
        )
    podman_args = datastores.podman_args(run)
    if size.extra["accelerators"]:
        image = driver.ex_get_image_from_family(
            "ubuntu-accelerator-2404-amd64-with-nvidia-570",
            ("ubuntu-os-accelerator-images",),
        )
        setup = textwrap.dedent(
            """\
            curl -fsSL https://nvidia.github.io/libnvidia-container/gpgkey | gpg --dearmor -o /usr/share/keyrings/nvidia-container-toolkit-keyring.gpg
            curl -s -L https://nvidia.github.io/libnvidia-container/stable/deb/nvidia-container-toolkit.list | \
                sed 's#deb https://#deb [signed-by=/usr/share/keyrings/nvidia-container-toolkit-keyring.gpg] https://#g' | \
                tee /etc/apt/sources.list.d/nvidia-container-toolkit.list
            apt-get -y update
            apt-get install -y \
                nvidia-container-toolkit \
                nvidia-container-toolkit-base \
                libnvidia-container-tools \
                libnvidia-container1
            nvidia-ctk cdi generate --output=/etc/cdi/nvidia.yaml
            """
        )
        podman_args += " --device nvidia.com/gpu=all"
        on_host_maintenance = "TERMINATE"
    else:
        image = driver.ex_get_image_from_family(
            "ubuntu-2404-lts-amd64", ("ubuntu-os-cloud",)
        )
        setup = ""
        on_host_maintenance = None

    teardown = textwrap.dedent(
        """\
        NAME=$(curl -X GET http://metadata.google.internal/computeMetadata/v1/instance/name -H 'Metadata-Flavor: Google')
        ZONE=$(curl -X GET http://metadata.google.internal/computeMetadata/v1/instance/zone -H 'Metadata-Flavor: Google')
        gcloud --quiet beta compute instances delete $NAME --zone=$ZONE --no-graceful-shutdown
        """
    )

    deployments = MultiStepDeployment(
        [
            services.mount_inputs(run.inputs),
            services.oo_context(run, key=key, setup=setup, teardown=teardown),
            services.job(
                command=safe_command(run),
                image=run.environment.image,
                podman_args=podman_args,
            ),
            services.start(),
        ]
    )

    driver.deploy_node(
        name=_node_name(run),
        ssh_key=config.settings().sshkey.as_posix(),
        size=size,
        image=image,
        ex_metadata=metadata,
        deploy=deployments,
        ex_service_accounts=[{"email": "default", "scopes": ["compute-rw"]}],
        ex_disk_size=int(run.environment.size.to("GB") + 0.5),
        ex_on_host_maintenance=on_host_maintenance,
    )


def _start_scaleway(run: Run, key: str):
    driver = get_driver(run.environment)
    sizes = driver.list_sizes(region=run.environment.region)
    size = next((s for s in sizes if s.name == run.environment.machinetype), None)
    if size is None:
        raise RuntimeError(
            f"{run.environment.machinetype} not available in {run.environment.region}"
        )

    # Get around libcloud checks
    size.extra["max_disk"] = run.environment.size.to("GB") + 21

    ex_volumes = {"0": {"size": run.environment.size, "volume_type": "sbs_volume"}}

    gpus = size.extra["gpus"] > 0
    podman_args = datastores.podman_args(run)
    DuckImage = collections.namedtuple("Image", ["id", "extra"])
    if gpus:
        image = DuckImage(extra={"size": 20}, id="ubuntu_noble_gpu_os_12")
        setup = "nvidia-ctk cdi generate --output=/etc/cdi/nvidia.yaml"
        podman_args += " --device nvidia.com/gpu=all"
    else:
        image = DuckImage(extra={"size": 10}, id="ubuntu_noble")
        setup = ""

    scaleway_key = config.scaleway_credentials()["SCW_SECRET_KEY"]
    teardown = textwrap.dedent(
        f"""\
        ID=$(scw-metadata | grep "^ID=" | cut -d "=" -f 2)
        ZONE=$(scw-metadata | grep "^ZONE=" | cut -d "=" -f 2)
        VOLUME=$(scw-metadata | grep "^VOLUMES_0_ID=" | cut -d "=" -f 2)
        curl --fail-early \
            -X POST \
            -H "X-Auth-Token: {scaleway_key}" \
            -H "Content-Type: application/json" \
            -d '{{"volume_id": "'$VOLUME'"}}' \
            https://api.scaleway.com/instance/v1/zones/$ZONE/servers/$ID/detach-volume \
          -: \
            -X DELETE \
            --retry 5 --retry-all-errors --fail \
            -H "X-Auth-Token: {scaleway_key}" \
            https://api.scaleway.com/block/v1/zones/$ZONE/volumes/$VOLUME \
          -: \
            -X POST \
            -H "X-Auth-Token: {scaleway_key}" \
            -H "Content-Type: application/json" \
            -d '{{"action":"terminate"}}' \
            https://api.scaleway.com/instance/v1/zones/$ZONE/servers/$ID/action
        """
    )

    deployments = MultiStepDeployment(
        [
            services.mount_inputs(run.inputs),
            services.oo_context(run, key=key, setup=setup, teardown=teardown),
            services.job(
                command=safe_command(run),
                image=run.environment.image,
                podman_args=podman_args,
            ),
            services.start(),
        ]
    )

    driver.deploy_node(
        name=_node_name(run),
        region=run.environment.region,
        ssh_key=config.settings().sshkey.as_posix(),
        size=size,
        image=image,
        ex_volumes=ex_volumes,
        deploy=deployments,
    )


class RunFinished(Exception):
    """Raised when run has completed the job"""

    def __init__(self, exit_code=0):
        super().__init__()
        self.exit_code = exit_code


def _log_messages(log, *, show_all=False):
    exit_pattern = re.compile("^o-o ExitStatus=(?P<exit_code>\w+)")
    exit_code = 0
    for line in log:
        entry = json.loads(line)
        # TODO: figure out how this is possible, sometimes getting list of ints
        if not isinstance(entry["MESSAGE"], str):
            continue

        exit_match = exit_pattern.match(entry["MESSAGE"])
        if exit_match:
            exit_code = exit_match.group("exit_code")
            match exit_match.group("exit_code"):
                case "TERM":
                    exit_code = 143
                case "KILL":
                    exit_code = 137
                case value:
                    exit_code = int(value)
            if not show_all:
                raise RunFinished(exit_code)

        if show_all or (
            entry.get("_SYSTEMD_UNIT", None) == "o-o-job.service"
            and entry.get("SYSLOG_IDENTIFIER", None) == "podman"
            and entry.get("_TRANSPORT", None) == "stdout"
        ):
            yield LogMessage(
                message=entry["MESSAGE"],
                timestamp=pendulum.from_timestamp(
                    int(entry["__REALTIME_TIMESTAMP"]) / 1000000
                ),
            )
    raise RunFinished(exit_code)


def _node_name(run: Run):
    return f"o-{run.sha}"


def _get_node(run: Run):
    driver = get_driver(run.environment)
    kwargs = {}
    if run.environment.provider == libcloud.compute.types.Provider.SCALEWAY:
        kwargs["region"] = run.environment.region

    node = next(
        (n for n in driver.list_nodes(**kwargs) if n.name == _node_name(run)),
        None,
    )
    return node


def _get_node_ip(run: Run):
    node = _get_node(run)
    if node is None:
        raise RuntimeError(f"{run.sha} not running")
    _, ip_addresses = node.driver.wait_until_running([node])[0]
    return ip_addresses[0]


def connect(run: Run, *, max_lines=200, show_all=False):
    """Connect to a job and yield output messages"""
    if run.ended is None:
        try:
            ip = _get_node_ip(run)
            with paramiko.SSHClient() as client:
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                client.connect(
                    ip,
                    username="root",
                    key_filename=config.settings().sshkey.as_posix(),
                )
                _, output, _ = client.exec_command(
                    f"journalctl --follow --lines={max_lines or 'all'} --output=json",
                )
                yield from _log_messages(output, show_all=show_all)

        except RunFinished:
            raise
        except Exception:
            try_datastore = True
        else:
            try_datastore = False
    else:
        try_datastore = True

    if try_datastore:
        output = datastores.get_logs(run)
        if max_lines is not None:
            output = output[-max_lines:]
        yield from _log_messages(output, show_all=show_all)


def stop(run):
    """Stop a running job, nicely"""
    ip = _get_node_ip(run)
    with paramiko.SSHClient() as client:
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(
            ip,
            username="root",
            key_filename=config.settings().sshkey.as_posix(),
        )
        client.exec_command("systemctl stop o-o-job")


def kill(run: Run):
    """Stop a running job by destroying the node

    NOTE: logs and outputs will likely be lost.
    """
    node = _get_node(run)

    if node is not None:
        node.destroy()
        if run.environment.provider == libcloud.compute.types.Provider.SCALEWAY:
            DuckStorageVolume = collections.namedtuple("StorageVolume", ["id"])

            # node could still be terminating and the volume in use, may need to retry
            destroy_volume_with_retry = libcloud.utils.retry.Retry(
                (libcloud.common.exceptions.BaseHTTPError,),
                retry_delay=5,
            )(node.driver.destroy_volume)

            for volume in node.extra["volumes"].values():
                destroy_volume_with_retry(
                    DuckStorageVolume(id=volume["id"]),
                    region=run.environment.region,
                )

    completed(run, exit_status=137)  # Bash exit code for sigkill


def completed(run: Run, *, exit_status=0):
    """Mark the run as completed, with given exit_status"""
    response = requests.put(
        f"{config.settings().apiurl}/runs/{run.project}/{run.sha}/completed",
        headers={"X-API-Key": config.settings().token},
        params={"exit_status": str(exit_status)},
    )
    if response.status_code == 422:
        raise RuntimeError(response.text)
    response.raise_for_status()


def read(project, run_sha):
    """Retrieve the run in the project with the given run_sha"""
    response = requests.get(
        f"{config.settings().apiurl}/runs/{project}/{run_sha}",
        headers={"X-API-Key": config.settings().token},
    )
    if response.status_code == 422:
        raise RuntimeError(response.text)
    response.raise_for_status()
    return Run.model_validate(response.json())


def read_all(project):
    """Retreive all runs in the project"""
    response = requests.get(
        f"{config.settings().apiurl}/runs/{project}",
        headers={"X-API-Key": config.settings().token},
    )
    response.raise_for_status()
    return [Run.model_validate(r) for r in response.json()]


def outputs(run: Run):
    """Retreive all the run's outputs"""
    return datastores.list_contents(run.datastore, f"{run.sha}/")


def wait_for_inputs(run: Run):
    """Wait for the run's inputs to complete successfully."""
    while True:
        errored = [
            i.short_sha
            for i in run.inputs
            if i.ended is not None and i.exit_status != 0
        ]
        if errored:
            raise RuntimeError(
                f"input(s) {errored} did not complete successfully, cannot start run"
            )

        unfinished = (i.ended is None for i in run.inputs)
        if not any(unfinished):
            break

        time.sleep(10)
        run = read(run.project, run.sha)
